/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.hk2.api;

import java.io.IOException;
import java.util.List;

import org.glassfish.hk2.utilities.DescriptorImpl;

/**
 * Implementations of this interface are used to populate HK2 
 * service locators from inhabitants files
 * 
 * @author jwells
 *
 */
public interface Populator {
    /**
     * This method can be used to populate the service locator with files that
     * have been written out using the {@link DescriptorImpl} writeObject method.
     * 
     * @param fileFinder An object that finds files in the environment.  If this is null
     * then the system will look in the service locator for an implementation of
     * DescriptorFileFinder.  If one is still not find this service will return an empty list
     * @param postProcessors post-processors that allows the environment to modify the set
     * of descriptors that are added to the system.
     * @return The list of descriptors added to the system.  Will not return null, but may return
     * an empty list
     * @throws IOException In case of an error reading the input streams
     * @throws MultiException if the user code throws an error, in which case none of the descriptors
     * will be added to the system
     */
    public List<ActiveDescriptor<?>> populate(
            DescriptorFileFinder fileFinder,
            PopulatorPostProcessor... postProcessors) throws IOException, MultiException;
    
    /**
     * This method will populate the service locator using the system classloader to
     * find the hk2-locator files from the default location of META-INF/hk2-locator/default.
     * No post processing will be done on the descriptors added to the system
     * 
     * @return The list of descriptors added to the system.  Will not return null, but may return
     * an empty list
     * @throws IOException if there was an error reading any of the descriptors
     * @throws MultiException if the user code throws an error, in which case none of the descriptors
     * will be added to the system
     */
    public List<ActiveDescriptor<?>> populate() throws IOException, MultiException; 

}
