/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.jvnet.hk2.internal;

import java.util.HashMap;
import java.util.LinkedList;

import org.glassfish.hk2.api.DescriptorVisibility;
import org.glassfish.hk2.api.Injectee;
import org.glassfish.hk2.api.InstantiationData;
import org.glassfish.hk2.api.InstantiationService;
import org.glassfish.hk2.api.Visibility;

/**
 * @author jwells
 *
 */
@Visibility(DescriptorVisibility.LOCAL)
public class InstantiationServiceImpl implements InstantiationService {
    private final HashMap<Long, LinkedList<Injectee>> injecteeStack = new HashMap<Long, LinkedList<Injectee>>();

    /* (non-Javadoc)
     * @see org.glassfish.hk2.api.InstantiationService#getInstantiationData()
     */
    @Override
    public synchronized InstantiationData getInstantiationData() {
        long tid = Thread.currentThread().getId();
        
        LinkedList<Injectee> threadStack = injecteeStack.get(tid);
        if (threadStack == null) return null;
        if (threadStack.isEmpty()) return null;
        
        final Injectee head = threadStack.getLast();
        
        return new InstantiationData() {

            @Override
            public Injectee getParentInjectee() {
                return head;
            }
            
            @Override
            public String toString() {
                return "InstantiationData(" + head + "," + System.identityHashCode(this) + ")";
            }
            
        };
   
    }
    
    public synchronized void pushInjecteeParent(Injectee injectee) {
        long tid = Thread.currentThread().getId();
        
        LinkedList<Injectee> threadStack = injecteeStack.get(tid);
        if (threadStack == null) {
            threadStack = new LinkedList<Injectee>();
            injecteeStack.put(tid, threadStack);
        }
        
        threadStack.addLast(injectee);
    }
    
    public synchronized void popInjecteeParent() {
        long tid = Thread.currentThread().getId();
        
        LinkedList<Injectee> threadStack = injecteeStack.get(tid);
        if (threadStack == null) return;
        
        threadStack.removeLast();
        
        if (threadStack.isEmpty()) {
            // prevents memory leaks for long dead threads
            injecteeStack.remove(tid);
        }
    }
    
    @Override
    public String toString() {
        return "InstantiationServiceImpl(" + injecteeStack.keySet() + "," + System.identityHashCode(this) + ")";
    }

}
