/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.hk2.tests.locator.singleton;

import java.util.LinkedList;
import java.util.List;

import junit.framework.Assert;

import org.glassfish.hk2.api.ActiveDescriptor;
import org.glassfish.hk2.api.ServiceHandle;
import org.glassfish.hk2.api.ServiceLocator;
import org.glassfish.hk2.api.ServiceLocatorFactory;
import org.glassfish.hk2.tests.locator.utilities.LocatorHelper;
import org.glassfish.hk2.utilities.BuilderHelper;
import org.junit.Test;

/**
 * 
 * @author jwells
 */
public class SingletonTest {
    private final static String TEST_NAME = "SingletonTest";
    private final static ServiceLocator locator = LocatorHelper.create(TEST_NAME, new SingletonModule());
    
    private final static String TEST_NAME2 = "SingletonTest2";
    private final static ServiceLocator locator2 = LocatorHelper.create(TEST_NAME2, new SingletonModule2());
    
    private final static int NUM_THREADS = 20;
    private final Object lock = new Object();
    private int threadsRun = 0;
    
    /**
     * The 100ms sleep in Single means that this test will fail nearly
     * every time if not protected by a single lock
     * 
     * @throws InterruptedException
     */
    @Test
    public void testOnlyOne() throws InterruptedException {
        List<Thread> serviceHandles = new LinkedList<Thread>();
        for (int lcv = 0; lcv < NUM_THREADS; lcv++) {
            ActiveDescriptor<?> ad = locator.getBestDescriptor(
                    BuilderHelper.createContractFilter(ConstructorCounterService.class.getName()));
            ServiceHandle<?> handle = locator.getServiceHandle(ad);
            Assert.assertNotNull(handle);
            
            serviceHandles.add(new Thread(new MyWorker(handle)));
        }
        
        for (Thread t : serviceHandles) {
            t.start();
        }
        
        synchronized (lock) {
            while (threadsRun < NUM_THREADS) {
                lock.wait();
            }
            
            Assert.assertEquals(threadsRun, NUM_THREADS);
        }
        
        Assert.assertEquals(ConstructorCounterService.getNumTimesInitialized(), 1);
    }
    
    /**
     * This tests that when a locator is shut down it will pre-destroy
     * singletons
     */
    @Test
    public void testLocatorShutDown() {
        Triple triple = locator2.getService(Triple.class);
        Assert.assertFalse(triple.getWasPreDestroyed());
        
        Double myDouble = locator2.getService(Double.class);
        Assert.assertFalse(myDouble.getWasPreDestroyed());
        
        // Shut down the locator
        ServiceLocatorFactory.getInstance().destroy(TEST_NAME2);
        
        Assert.assertTrue(myDouble.getWasPreDestroyed());
        Assert.assertTrue(triple.getWasPreDestroyed());
    }
    
    private class MyWorker implements Runnable {
        private final ServiceHandle<?> handle;
        
        private MyWorker(ServiceHandle<?> handle) {
            this.handle = handle;
        }

        @Override
        public void run() {
            handle.getService();
            
            synchronized (lock) {
                threadsRun++;
                if (threadsRun >= NUM_THREADS) {
                    lock.notify();
                }
            }
        }
        
    }

}
