/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.hk2.tests.locator.defaultunqualified;

import org.glassfish.hk2.api.IterableProvider;
import org.glassfish.hk2.api.ServiceLocator;
import org.glassfish.hk2.tests.locator.utilities.LocatorHelper;
import org.glassfish.hk2.utilities.UnqualifiedImpl;
import org.junit.Assert;
import org.junit.Test;

/**
 * @author jwells
 *
 */
public class DefaultUnqualifiedTest {
    /**
     * Tests that when using a system default unqualified
     * that the service gets the proper non-qualified
     * service.  Note that DollImpl has a low priority, and
     * hence would NOT be picked for the doll injection
     * point if the DefaultUnqualified were not set
     */
    @Test // @org.junit.Ignore
    public void testGetsUnqualifiedService() {
        ServiceLocator locator = LocatorHelper.getServiceLocator(BoardGameImpl.class,
                DollImpl.class,
                TrainImpl.class,
                ChildsRoom.class);
        
        locator.setDefaultUnqualified(new UnqualifiedImpl());
        
        ChildsRoom childsRoom = locator.getService(ChildsRoom.class);
        
        Assert.assertEquals(BoardGameImpl.class, childsRoom.getBoardGame().getClass());
        Assert.assertEquals(TrainImpl.class, childsRoom.getTrain().getClass());
        
        // The real test:
        Assert.assertEquals(DollImpl.class, childsRoom.getDoll().getClass());
        Assert.assertEquals(DollImpl.class, childsRoom.getDollProvider().get().getClass());
    }
    
    /**
     * Tests that injection of IterableProvider does NOT have the
     * default Unqualified applied to it
     */
    @Test // @org.junit.Ignore
    public void testIterableProviderReturnsAllEvenIfUnqualified() {
        ServiceLocator locator = LocatorHelper.getServiceLocator(BoardGameImpl.class,
                DollImpl.class,
                TrainImpl.class,
                ChildsRoom.class);
        
        locator.setDefaultUnqualified(new UnqualifiedImpl());
        
        ChildsRoom childsRoom = locator.getService(ChildsRoom.class);
        
        IterableProvider<Toy> allToys = childsRoom.getAllToys();
        Assert.assertEquals(3, allToys.getSize());
        
        int lcv = 0;
        for (Toy toy : allToys) {
            switch (lcv) {
            case 0:
                Assert.assertEquals(BoardGameImpl.class, toy.getClass());
                break;
            case 1:
                Assert.assertEquals(TrainImpl.class, toy.getClass());
                break;
            case 2:
                Assert.assertEquals(DollImpl.class, toy.getClass());
                break;
            default:
                Assert.fail("Should not get here");
            }
            
            lcv++;
        }
        
        
    }

}
