/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2007-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.jvnet.hk2.config;

import org.jvnet.hk2.annotations.Service;

/** Represents a Java primitive (and its wrapper) data type. Not all Java primitives
 *  are relevant from a configuration standpoint.
 * @see DataType
 * @see DomDocument#PRIMS
 * @author &#2325;&#2375;&#2342;&#2366;&#2352 (km@dev.java.net)
 * @since hk2 0.3.10
 */
public final class PrimitiveDataType implements DataType {

    private final String realType;
    PrimitiveDataType(String realType) {
        assert DomDocument.PRIMS.contains(realType) : "This class can't validate: " + realType;
        this.realType = realType;
    }
    
    public void validate(String value) throws ValidationException {
        if (value.startsWith("${") && value.endsWith("}")) //it's a token
          return;
        boolean match = false;
        if ("int".equals(realType) || "java.lang.Integer".equals(realType))
            match = representsInteger(value);
        else if ("boolean".equals(realType) || "java.lang.Boolean".endsWith(realType))
            match = representsBoolean(value);
        else if ("char".equals(realType) || "java.lang.Character".equals(realType))
            representsChar(value);
        //no need for else as we are asserting it in the constructor
        if (!match) {
            String msg = "This value: " + value + " is not of type: " + realType + ", validation failed";
            throw new ValidationException(msg);            
        }
    }
    
    private static boolean representsBoolean(String value) {
        boolean isBoolean = "true".equalsIgnoreCase(value) || "false".equalsIgnoreCase(value);
        return (isBoolean);
    }
    private static boolean representsChar(String value) {
            if (value.length() == 1)
                return true;
            return false;
    }
    private static boolean representsInteger(String value) {
        try {
            Integer.parseInt(value);
            return true;
        } catch(NumberFormatException ne) {
            return false;
        }
    }

}
