/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.hk2.configuration.hub.xml.dom.integration;

import org.jvnet.hk2.annotations.Contract;

/**
 * This service is used to translate hk2-config bean data from
 * hk2-config to the {@link Hub}.  This integration layer has
 * made several choices about type name (x-path) and instance
 * name (x-path like, separated by dots with non-keyed beans
 * taking the name of the xml attribute).  It may be that another
 * system would like to choose different names, or may even
 * which to transform the bean into a different form.  This
 * service allows the user of this integration layer to map
 * the system chosen naming scheme to one of their own liking
 * 
 * @author jwells
 *
 */
@Contract
public interface XmlDomTranslationService {
    /**
     * This method is called by the system prior to adding, modifying
     * or removing an hk2-bean.  If this method returns non-null, then
     * the data in the returned {@link XmlDomHubData} will be used when
     * interacting with the {@link Hub}.
     * <p>
     * If multiple services exist they will all be called in priority
     * order, with the results from one call being passed into the
     * parameter of the next call
     * 
     * @param hk2ConfigBeanData The never null information that will be
     * used with the {@link Hub} when adding, modifying or removing hk2-config
     * beans
     * @return if this method returns null then the incoming data will not
     * be modified.  If this method returns non-null then the data used
     * in the Hub will be from the returned value
     */
    public XmlDomHubData translate(XmlDomHubData hk2ConfigBeanData);

}
