/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2012-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.jvnet.hk2.testing.junit.internal;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;

import java.util.Collection;
import java.util.Collections;

import javax.inject.Inject;
import javax.inject.Singleton;

import org.glassfish.hk2.api.ActiveDescriptor;
import org.glassfish.hk2.api.DynamicConfiguration;
import org.glassfish.hk2.api.DynamicConfigurationService;
import org.glassfish.hk2.api.Injectee;
import org.glassfish.hk2.api.JustInTimeInjectionResolver;

/**
 * @author jwells
 *
 */
@Singleton
public class JustInTimeInjectionResolverImpl implements
        JustInTimeInjectionResolver {

    private final Collection<?> excludes;
    
    @Inject
    private DynamicConfigurationService dcs;

    public JustInTimeInjectionResolverImpl() {
        this(Collections.emptySet());
    }

    public JustInTimeInjectionResolverImpl(final Collection<?> excludes) {
        super();
        if (excludes == null) {
            this.excludes = Collections.emptySet();
        } else {
            this.excludes = excludes;
        }
    }

    /**
     * Returns {@code true} if the supplied {@link Injectee}
     * represents a service lookup rather than a true injection point.
     *
     * @param failedInjectionPoint the {@link Injectee} to test; may
     * be {@code null} in which case {@code true} will be returned
     *
     * @return {@code true} if the supplied {@link Injectee}
     * represents a service lookup rather than a true injection point
     */
    protected boolean isLookup(final Injectee failedInjectionPoint) {
        return failedInjectionPoint == null || failedInjectionPoint.getParent() == null;
    }
  
    /* (non-Javadoc)
     * @see org.glassfish.hk2.api.JustInTimeInjectionResolver#justInTimeResolution(org.glassfish.hk2.api.Injectee)
     */
    @Override
    public boolean justInTimeResolution(Injectee failedInjectionPoint) {
        boolean returnValue = false;
        if (failedInjectionPoint != null && !isLookup(failedInjectionPoint)) {
            
            Type needType = failedInjectionPoint.getRequiredType();
            
            Class<?> needClass = null;
            if (needType instanceof Class) {
                needClass = (Class<?>) needType;
            }
            else if (needType instanceof ParameterizedType) {
                Type rawType = ((ParameterizedType) needType).getRawType();
                if (rawType instanceof Class) {
                    needClass = (Class<?>) rawType;
                }
            }
            
            if (needClass == null || needClass.isInterface() || (excludes != null && excludes.contains(needClass.getName()))) {
                return false;
            }
            
            DynamicConfiguration config = dcs.createDynamicConfiguration();
            
            try {
                final ActiveDescriptor<?> ad = config.addActiveDescriptor(needClass);
                config.commit();
                returnValue = true;
            }
            catch (Throwable th) {
                returnValue = false;
            }
        }
        return returnValue;
    }

}
