/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.hk2.tests.locator.validating;

import javax.inject.Singleton;

import org.glassfish.hk2.api.Filter;
import org.glassfish.hk2.api.ValidationInformation;
import org.glassfish.hk2.api.ValidationService;
import org.glassfish.hk2.api.Validator;
import org.glassfish.hk2.utilities.BuilderHelper;

/**
 * @author jwells
 *
 */
@Singleton
public class InvalidCheckCallerValidationService implements ValidationService {
    private final MyValidator validator = new MyValidator();

    @Override
    public Filter getLookupFilter() {
        return BuilderHelper.allFilter();
    }

    @Override
    public Validator getValidator() {
        return validator;
    }
    
    public boolean check() {
        return validator.check();
    }
    
    public static class MyValidator implements Validator {
        private boolean done = false;
        private boolean check = false;

        @Override
        public boolean validate(ValidationInformation info) {
            MyRunner runner = new MyRunner(this, info);
            
            Thread t = new Thread(runner);
            t.start();
            
            synchronized (this) {
                while (!done) {
                    try {
                        this.wait();
                    }
                    catch (InterruptedException e) {
                        throw new RuntimeException(e);
                    }
                }
            }
            
            check = (runner.getElement() == null);
            
            return true;
        }
        
        public void done() {
            done = true;
        }
        
        public boolean check() {
            return check;
        }
        
    }
    
    public static class MyRunner implements Runnable {
        private final MyValidator parent;
        private final ValidationInformation info;
        private StackTraceElement element;
        
        private MyRunner(MyValidator parent, ValidationInformation info) {
            this.parent = parent;
            this.info = info;
        }

        @Override
        public void run() {
            synchronized (parent) {
                element = info.getCaller();
                parent.done();
                
                parent.notify();
            }
        }
        
        public StackTraceElement getElement() {
            synchronized (parent) {
                return element;
            }
        }
    }

}
