/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2014 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.hk2.tests.locator.context.ghost;

import java.lang.annotation.Annotation;

import javax.inject.Singleton;

import org.glassfish.hk2.api.ActiveDescriptor;
import org.glassfish.hk2.api.AnnotationLiteral;
import org.glassfish.hk2.api.Descriptor;
import org.glassfish.hk2.api.PerLookup;
import org.glassfish.hk2.api.ServiceLocator;
import org.glassfish.hk2.tests.locator.utilities.LocatorHelper;
import org.glassfish.hk2.utilities.BuilderHelper;
import org.glassfish.hk2.utilities.ServiceLocatorUtilities;
import org.junit.Assert;
import org.junit.Test;

/**
 * @author jwells
 *
 */
public class ContextGhostTest {
    /**
     * Ensures that you can change from one hard-coded scope on the descriptor
     * to a different scope
     */
    @Test
    public void testSwitchFromExplicitScopeToGhostedScope() {
        ServiceLocator locator = LocatorHelper.create();
        ServiceLocatorUtilities.enableLookupExceptions(locator);
        ServiceLocatorUtilities.addClasses(locator, GhostedContext.class);
        
        Descriptor desc = BuilderHelper.activeLink(SingletonScopedService.class).
                to(SingletonScopedService.class).
                in(new GhostedScopeImpl(0)).
                build();
        
        ServiceLocatorUtilities.addOneDescriptor(locator, desc);
        
        SingletonScopedService one = locator.getService(SingletonScopedService.class);
        SingletonScopedService two = locator.getService(SingletonScopedService.class);
        
        Assert.assertNotNull(one);
        Assert.assertNotSame(one, two);
    }
    
    /**
     * Ensures we can add a scope to a service with no scope at all
     */
    @Test
    public void testGiveClassWithNoScopeAScope() {
        ServiceLocator locator = LocatorHelper.create();
        ServiceLocatorUtilities.enableLookupExceptions(locator);
        
        Descriptor desc = BuilderHelper.activeLink(NoScopeService.class).
                to(NoScopeService.class).
                in(ServiceLocatorUtilities.getSingletonAnnotation()).
                build();
        
        ServiceLocatorUtilities.addOneDescriptor(locator, desc);
        
        NoScopeService one = locator.getService(NoScopeService.class);
        NoScopeService two = locator.getService(NoScopeService.class);
        
        Assert.assertNotNull(one);
        Assert.assertEquals(one, two);
        
    }
    
    /**
     * Tests that we can change the value of a field in a
     * scope with a ghost added annotation
     */
    @Test
    public void testModifyExistingScopeWithDifferentValue() {
        ServiceLocator locator = LocatorHelper.create();
        ServiceLocatorUtilities.addClasses(locator, GhostedContext.class);
        
        Descriptor desc = BuilderHelper.activeLink(GhostedServiceWithValue.class).
                to(GhostedServiceWithValue.class).
                in(new GhostedScopeImpl(0)).
                build();
        
        ServiceLocatorUtilities.addOneDescriptor(locator, desc);
        
        GhostedServiceWithValue ghosted = locator.getService(GhostedServiceWithValue.class);
        Assert.assertNotNull(ghosted);
        
        ActiveDescriptor<?> gDesck = ghosted.getDescriptor();
        
        Annotation anno = gDesck.getScopeAsAnnotation();
        Assert.assertNotNull(anno);
        
        GhostedScope gs = (GhostedScope) anno;
        Assert.assertEquals(0, gs.value());
        
    }
    
    private static class GhostedScopeImpl extends AnnotationLiteral<GhostedScope> implements GhostedScope {
        private final int value;
        
        private GhostedScopeImpl(int value) {
            this.value = value;
        }

        /* (non-Javadoc)
         * @see org.glassfish.hk2.tests.locator.context.ghost.GhostedScope#value()
         */
        @Override
        public int value() {
            return value;
        }
    }

}
