/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2014-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.hk2.tests.locator.messaging.error;

import javax.inject.Singleton;

import org.glassfish.hk2.api.MultiException;
import org.glassfish.hk2.api.ServiceLocator;
import org.glassfish.hk2.api.messaging.Topic;
import org.glassfish.hk2.extras.events.DefaultTopicDistributionErrorService;
import org.glassfish.hk2.tests.extras.internal.Utilities;
import org.glassfish.hk2.utilities.ServiceLocatorUtilities;
import org.junit.Assert;
import org.junit.Test;

/**
 * Tests various failure scenarios
 * 
 * @author jwells
 */
public class TopicErrorCasesTest {
    public final static String EXPECTED_MESSAGE = "ExpectedMessage";
    
    /**
     * Tests that a subscribers exception does not stop another
     * subscriber from getting the message
     */
    @Test
    public void testSubscriberThrows() {
        ServiceLocator locator = Utilities.getLocatorWithTopics();
        
        ServiceLocatorUtilities.addClasses(locator, SubscriberThrowsException.class,
                Subscriber.class,
                Publisher.class,
                ErrorHandler.class);
        
        locator.getService(SubscriberThrowsException.class);
        Subscriber subscriber = locator.getService(Subscriber.class);
        Publisher publisher = locator.getService(Publisher.class);
        ErrorHandler errorHandler = locator.getService(ErrorHandler.class);
        
        MyEvent event = new MyEvent();
        publisher.publish(event);
        
        // Ensures that other subscribers get the event
        Assert.assertEquals(1, subscriber.getNumEvents());
        
        MultiException me = errorHandler.lastError;
        Assert.assertNotNull(me);
        
        Assert.assertTrue(me.toString().contains(EXPECTED_MESSAGE));
        
        Assert.assertEquals(event, errorHandler.lastMessage);
        Assert.assertEquals(publisher.getTopic(), errorHandler.lastTopic);
    }
    
    @Singleton
    private static class ErrorHandler implements DefaultTopicDistributionErrorService {
        private Topic<?> lastTopic;
        private Object lastMessage;
        private MultiException lastError;

        @Override
        public void subscribersFailed(Topic<?> topic, Object message,
                MultiException error) {
            lastTopic = topic;
            lastMessage = message;
            lastError = error;
        }
        
    }

}
