/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2014-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */
package org.glassfish.hk2.testing.collections;

import java.util.Collection;
import java.util.Set;

/**
 * This class can be used to test Sets for basic
 * Set functionality.
 * <p>
 * Subclasses must implement the methods that return
 * Sets from the two most common types of constructors,
 * the zero argument constructor and the single argument
 * Collection constructor.
 * <p>
 * In general, sets do not allow for duplicate entries, but
 * this may not be true of all sets.  All of the same questions
 * about Collections can also be overriden for Sets
 */
@SuppressWarnings({"rawtypes"})
public abstract class AbstractSetTest extends AbstractCollectionTest {
  /**
   * Subclasses should override this method if their Set
   * implementation allows duplicate elements.  In other words,
   * if you call "add(foo)" followed by "add(foo)" and the Set
   * will have two elements then your Set supports duplicate
   * elements.  Otherwise (if duplicates are overwritten) then this
   * method should return false.
   * 
   * @return true if the set allows an arbitrary number of
   * elements whose "equals" returns true to be in the collection, and
   * false if those objects will instead be overwritten
   */
  protected boolean doesCollectionSupportDuplicateElements() {
    return false;
  }

  /* (non-Javadoc)
   * @see com.oracle.weblogic.testing.collections.AbstractCollectionTest#createCollection()
   */
  @Override
  protected Collection createCollection() {
    return createSet();
  }

  /* (non-Javadoc)
   * @see com.oracle.weblogic.testing.collections.AbstractCollectionTest#createCollection(java.util.Collection)
   */
  @Override
  protected Collection createCollection(Collection input) {
    return createSet(input);
  }
  
  protected abstract Set createSet();
  
  protected abstract Set createSet(Collection input);

}
