/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright (c) 2013-2015 Oracle and/or its affiliates. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common Development
 * and Distribution License("CDDL") (collectively, the "License").  You
 * may not use this file except in compliance with the License.  You can
 * obtain a copy of the License at
 * https://glassfish.dev.java.net/public/CDDL+GPL_1_1.html
 * or packager/legal/LICENSE.txt.  See the License for the specific
 * language governing permissions and limitations under the License.
 *
 * When distributing the software, include this License Header Notice in each
 * file and include the License file at packager/legal/LICENSE.txt.
 *
 * GPL Classpath Exception:
 * Oracle designates this particular file as subject to the "Classpath"
 * exception as provided by Oracle in the GPL Version 2 section of the License
 * file that accompanied this code.
 *
 * Modifications:
 * If applicable, add the following below the License Header, with the fields
 * enclosed by brackets [] replaced by your own identifying information:
 * "Portions Copyright [year] [name of copyright owner]"
 *
 * Contributor(s):
 * If you wish your version of this file to be governed by only the CDDL or
 * only the GPL Version 2, indicate your decision by adding "[Contributor]
 * elects to include this software in this distribution under the [CDDL or GPL
 * Version 2] license."  If you don't indicate a single choice of license, a
 * recipient has the option to distribute your version of this file under
 * either the CDDL, the GPL Version 2 or to extend the choice of license to
 * its licensees as provided above.  However, if you add GPL Version 2 code
 * and therefore, elected the GPL Version 2 license, then the option applies
 * only if the new code is made subject to such option by the copyright
 * holder.
 */

package org.glassfish.hk2.runlevel.tests.executor;

import java.util.concurrent.Executor;

import org.glassfish.hk2.api.ServiceLocator;
import org.glassfish.hk2.runlevel.RunLevelController;
import org.glassfish.hk2.runlevel.tests.utilities.Utilities;
import org.junit.Assert;
import org.junit.Test;

/**
 * @author jwells
 *
 */
public class ExecutorTest {
    @Test
    public void testCustomExecutor() {
        ServiceLocator locator = Utilities.getServiceLocator(
                ThreadNameRecorderService.class);
        
        RunLevelController controller = locator.getService(RunLevelController.class);
        
        controller.proceedTo(5);
        
        ThreadNameRecorderService service = locator.getService(ThreadNameRecorderService.class);
        Assert.assertFalse(ExecutorImpl.THREAD_NAME.equals(service.getThreadName()));
        
        controller.proceedTo(4);
        
        controller.setExecutor(new ExecutorImpl());
        
        controller.proceedTo(5);
        
        service = locator.getService(ThreadNameRecorderService.class);
        Assert.assertTrue(ExecutorImpl.THREAD_NAME.equals(service.getThreadName()));
        
        controller.proceedTo(4);
        
        controller.setExecutor(null);
        
        controller.proceedTo(5);
        
        service = locator.getService(ThreadNameRecorderService.class);
        Assert.assertFalse(ExecutorImpl.THREAD_NAME.equals(service.getThreadName()));
    }
    
    @Test
    public void testGetExecutor() {
        ServiceLocator locator = Utilities.getServiceLocator(
                ThreadNameRecorderService.class);
        
        RunLevelController controller = locator.getService(RunLevelController.class);
        
        Executor defaultExecutor = controller.getExecutor();
        Assert.assertNotNull(defaultExecutor);
        
        Executor myExecutor = new ExecutorImpl();
        
        controller.setExecutor(myExecutor);
        
        Assert.assertEquals(myExecutor, controller.getExecutor());
        
        Executor secondExecutor = new ExecutorImpl();
        
        controller.setExecutor(secondExecutor);
        
        Assert.assertEquals(secondExecutor, controller.getExecutor());
        
        controller.setExecutor(null);
        
        Assert.assertEquals(defaultExecutor, controller.getExecutor());
    }
    
    public static class ExecutorImpl implements Executor {
        public final static String THREAD_NAME = "TestThread";

        @Override
        public void execute(Runnable command) {
            Thread t = new Thread(command);
            t.setDaemon(true);
            t.setName(THREAD_NAME);
            
            t.start();
        }
        
    }

}
